@@ -1,5 +1,5 @@
 // Copyright (C) Andrew Tridgell 2002 (original file)
-// Copyright (C) 2006-2007 Red Hat Inc. (systemtap changes)
+// Copyright (C) 2006-2008 Red Hat Inc. (systemtap changes)
 // 
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -65,9 +65,6 @@ hash::result(string& r)
 }
 
 
-// Grabbed from linux/module.h kernel include.
-#define MODULE_NAME_LEN (64 - sizeof(unsigned long))
-
 void
 find_hash (systemtap_session& s, const string& script)
 {
@@ -100,10 +97,32 @@ find_hash (systemtap_session& s, const string& script)
   h.add(s.merge);			// '-M'
   h.add(s.timing);			// '-t'
   h.add(s.prologue_searching);		// '-P'
+  h.add(s.ignore_vmlinux);		// --ignore-vmlinux
+  h.add(s.ignore_dwarf);		// --ignore-dwarf
+  h.add(s.consult_symtab);		// --kelf, --kmap
+  if (!s.kernel_symtab_path.empty())	// --kmap
+    {
+      h.add(s.kernel_symtab_path);
+      if (stat(s.kernel_symtab_path.c_str(), &st) == 0)
+        {
+	  // NB: stat of /proc/kallsyms always returns size=0, mtime=now...
+	  // which is a good reason to use the default /boot/System.map-2.6.xx
+	  // instead.
+          h.add(st.st_size);
+	  h.add(st.st_mtime);
+        }
+    }
   for (unsigned i = 0; i < s.macros.size(); i++)
     h.add(s.macros[i]);
 
-  // Hash runtime path (that gets added in as "-I path").
+  // -d MODULE
+  for (set<string>::iterator it = s.unwindsym_modules.begin();
+       it != s.unwindsym_modules.end();
+       it++)
+    h.add(*it);
+    // XXX: a build-id of each module might be even better
+
+  // Hash runtime path (that gets added in as "-R path").
   h.add(s.runtime_path);
 
   // Hash compiler path, size, and mtime.  We're just going to assume
