@@ -19,12 +19,14 @@
 #include "cache.h"
 #include "util.h"
 #include "coveragedb.h"
+#include "git_version.h"
 
 #include <iostream>
 #include <fstream>
 #include <sstream>
 #include <cerrno>
 #include <cstdlib>
+#include <limits.h>
 
 extern "C" {
 #include <glob.h>
@@ -36,6 +38,7 @@ extern "C" {
 #include <sys/stat.h>
 #include <time.h>
 #include <elfutils/libdwfl.h>
+#include <getopt.h>
 }
 
 using namespace std;
@@ -47,7 +50,7 @@ version ()
   clog
     << "SystemTap translator/driver "
     << "(version " << VERSION << "/" << dwfl_version (NULL)
-    << " built " << DATE << ")" << endl
+    << " " << GIT_MESSAGE << ")" << endl
     << "Copyright (C) 2005-2008 Red Hat, Inc. and others" << endl
     << "This is free software; see the source for copying conditions." << endl;
 }
@@ -64,9 +67,11 @@ usage (systemtap_session& s, int exitcode)
     << endl
     << "   or: stap [options] -e SCRIPT    Run given script."
     << endl
+    << "   or: stap [options] -l PROBE     List matching probes."
+    << endl
     << endl
     << "Options:" << endl
-    << "   --         no more options after this" << endl
+    << "   --         end of translator options, script options follow" << endl
     << "   -v         increase verbosity [" << s.verbose << "]" << endl
     << "   -h         show help" << endl
     << "   -V         show version" << endl
@@ -101,10 +106,28 @@ usage (systemtap_session& s, int exitcode)
     << "   -c CMD     start the probes, run CMD, and exit when it finishes"
     << endl
     << "   -x PID     sets target() to PID" << endl
+    << "   -d OBJECT  add unwind/symbol data for OBJECT file";
+  if (s.unwindsym_modules.size() == 0)
+    clog << endl;
+  else
+    clog << ", in addition to" << endl;
+  {
+    vector<string> syms (s.unwindsym_modules.begin(), s.unwindsym_modules.end());
+    for (unsigned i=0; i<syms.size(); i++)
+      clog << "              " << syms[i] << endl;
+  }
+  clog
     << "   -t         collect probe timing information" << endl
 #ifdef HAVE_LIBSQLITE3
-    << "   -q         generate information on tapset coverage"
+    << "   -q         generate information on tapset coverage" << endl
 #endif /* HAVE_LIBSQLITE3 */
+    << "   --kelf     make do with symbol table from vmlinux" << endl
+    << "   --kmap[=FILE]" << endl
+    << "              make do with symbol table from nm listing" << endl
+    << "   --ignore-vmlinux" << endl
+    << "              for testing, pretend vmlinux can't be found" << endl
+    << "   --ignore-dwarf" << endl
+    << "              for testing, pretend vmlinux and modules lack debug info"
     << endl
     ;
   // -d: dump safety-related external references
@@ -116,73 +139,120 @@ usage (systemtap_session& s, int exitcode)
 static void
 printscript(systemtap_session& s, ostream& o)
 {
-  if (s.embeds.size() > 0)
-    o << "# global embedded code" << endl;
-  for (unsigned i=0; i<s.embeds.size(); i++)
+  if (s.listing_mode)
     {
-      embeddedcode* ec = s.embeds[i];
-      ec->print (o);
-      o << endl;
-    }
+      // We go through some heroic measures to produce clean output.
+      set<string> seen;
 
-  if (s.globals.size() > 0)
-    o << "# globals" << endl;
-  for (unsigned i=0; i<s.globals.size(); i++)
-    {
-      vardecl* v = s.globals[i];
-      v->printsig (o);
-      if (s.verbose && v->init)
+      for (unsigned i=0; i<s.probes.size(); i++)
         {
-          o << " = ";
-          v->init->print(o);
+          derived_probe* p = s.probes[i];
+          // NB: p->basest() is not so interesting;
+          // p->almost_basest() doesn't quite work, so ...
+          vector<probe*> chain;
+          p->collect_derivation_chain (chain);
+          probe* second = (chain.size()>1) ? chain[chain.size()-2] : chain[0];
+
+          #if 0
+          cerr << "\tchain[" << chain.size() << "]:" << endl;
+          for (unsigned i=0; i<chain.size(); i++)
+            { cerr << "\t"; chain[i]->printsig(cerr); cerr << endl; }
+          #endif
+
+          stringstream tmps;
+          const probe_alias *a = second->get_alias();
+          if (a)
+            {
+              assert (a->alias_names.size() >= 1);
+              a->alias_names[0]->print(tmps); // XXX: [0] is arbitrary; perhaps print all
+            }            
+          else
+            {
+              assert (second->locations.size() >= 1);
+              second->locations[0]->print(tmps); // XXX: [0] is less arbitrary here, but still ...
+            }
+          string pp = tmps.str();
+
+          // Now duplicate-eliminate.  An alias may have expanded to
+          // several actual derived probe points, but we only want to
+          // print the alias head name once.
+          if (seen.find (pp) == seen.end())
+            {
+              o << pp << endl;
+              seen.insert (pp);
+            }
         }
-      o << endl;
     }
-
-  if (s.functions.size() > 0)
-    o << "# functions" << endl;
-  for (unsigned i=0; i<s.functions.size(); i++)
+  else
     {
-      functiondecl* f = s.functions[i];
-      f->printsig (o);
-      o << endl;
-      if (f->locals.size() > 0)
-	o << "  # locals" << endl;
-      for (unsigned j=0; j<f->locals.size(); j++)
+      if (s.embeds.size() > 0)
+        o << "# global embedded code" << endl;
+      for (unsigned i=0; i<s.embeds.size(); i++)
         {
-	  vardecl* v = f->locals[j];
-	  o << "  ";
-	  v->printsig (o);
-	  o << endl;
-	}
-      if (s.verbose)
+          embeddedcode* ec = s.embeds[i];
+          ec->print (o);
+          o << endl;
+        }
+      
+      if (s.globals.size() > 0)
+        o << "# globals" << endl;
+      for (unsigned i=0; i<s.globals.size(); i++)
         {
-	  f->body->print (o);
-	  o << endl;
-	}
-    }
-
-  if (s.probes.size() > 0)
-    o << "# probes" << endl;
-  for (unsigned i=0; i<s.probes.size(); i++)
-    {
-      derived_probe* p = s.probes[i];
-      p->printsig (o);
-      o << endl;
-      if (p->locals.size() > 0)
-        o << "  # locals" << endl;
-      for (unsigned j=0; j<p->locals.size(); j++)
+          vardecl* v = s.globals[i];
+          v->printsig (o);
+          if (s.verbose && v->init)
+            {
+              o << " = ";
+              v->init->print(o);
+            }
+          o << endl;
+        }
+      
+      if (s.functions.size() > 0)
+        o << "# functions" << endl;
+      for (unsigned i=0; i<s.functions.size(); i++)
         {
-	  vardecl* v = p->locals[j];
-	  o << "  ";
-	  v->printsig (o);
-	  o << endl;
-	}
-      if (s.verbose)
+          functiondecl* f = s.functions[i];
+          f->printsig (o);
+          o << endl;
+          if (f->locals.size() > 0)
+            o << "  # locals" << endl;
+          for (unsigned j=0; j<f->locals.size(); j++)
+            {
+              vardecl* v = f->locals[j];
+              o << "  ";
+              v->printsig (o);
+              o << endl;
+            }
+          if (s.verbose)
+            {
+              f->body->print (o);
+              o << endl;
+            }
+        }
+      
+      if (s.probes.size() > 0)
+        o << "# probes" << endl;
+      for (unsigned i=0; i<s.probes.size(); i++)
         {
-	  p->body->print (o);
-	  o << endl;
-	}
+          derived_probe* p = s.probes[i];
+          p->printsig (o);
+          o << endl;
+          if (p->locals.size() > 0)
+            o << "  # locals" << endl;
+          for (unsigned j=0; j<p->locals.size(); j++)
+            {
+              vardecl* v = p->locals[j];
+              o << "  ";
+              v->printsig (o);
+              o << endl;
+            }
+          if (s.verbose)
+            {
+              p->body->print (o);
+              o << endl;
+            }
+        }
     }
 }
 
@@ -203,6 +273,29 @@ void handle_interrupt (int /* sig */)
 }
 
 
+void
+setup_signals (sighandler_t handler)
+{
+  struct sigaction sa;
+
+  sa.sa_handler = handler;
+  sigemptyset (&sa.sa_mask);
+  if (handler != SIG_IGN)
+    {
+      sigaddset (&sa.sa_mask, SIGHUP);
+      sigaddset (&sa.sa_mask, SIGPIPE);
+      sigaddset (&sa.sa_mask, SIGINT);
+      sigaddset (&sa.sa_mask, SIGTERM);
+    }
+  sa.sa_flags = 0;
+
+  sigaction (SIGHUP, &sa, NULL);
+  sigaction (SIGPIPE, &sa, NULL);
+  sigaction (SIGINT, &sa, NULL);
+  sigaction (SIGTERM, &sa, NULL);
+}
+
+
 int
 main (int argc, char * const argv [])
 {
@@ -224,6 +317,7 @@ main (int argc, char * const argv [])
   s.bulk_mode = false;
   s.unoptimized = false;
   s.suppress_warnings = false;
+  s.listing_mode = false;
 
 #ifdef ENABLE_PROLOGUES
   s.prologue_searching = true;
@@ -244,6 +338,9 @@ main (int argc, char * const argv [])
   s.use_cache = true;
   s.tapset_compile_coverage = false;
   s.need_uprobes = false;
+  s.consult_symtab = false;
+  s.ignore_vmlinux = false;
+  s.ignore_dwarf = false;
 
   const char* s_p = getenv ("SYSTEMTAP_TAPSET");
   if (s_p != NULL)  
@@ -294,8 +391,21 @@ main (int argc, char * const argv [])
 
   while (true)
     {
+      int long_opt;
+#define LONG_OPT_KELF 1
+#define LONG_OPT_KMAP 2
+#define LONG_OPT_IGNORE_VMLINUX 3
+#define LONG_OPT_IGNORE_DWARF 4
       // NB: also see find_hash(), usage(), switch stmt below, stap.1 man page
-      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:uqw");
+      static struct option long_options[] = {
+        { "kelf", 0, &long_opt, LONG_OPT_KELF },
+        { "kmap", 2, &long_opt, LONG_OPT_KMAP },
+        { "ignore-vmlinux", 0, &long_opt, LONG_OPT_IGNORE_VMLINUX },
+        { "ignore-dwarf", 0, &long_opt, LONG_OPT_IGNORE_DWARF },
+        { NULL, 0, NULL, 0 }
+      };
+      int grc = getopt_long (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:uqwl:d:",
+                                                          long_options, NULL);
       if (grc < 0)
         break;
       switch (grc)
@@ -321,6 +431,11 @@ main (int argc, char * const argv [])
 	  break;
 
         case 'p':
+          if (s.listing_mode)
+            {
+              cerr << "Listing (-l) mode implies pass 2." << endl;
+              usage (s, 1);
+            }
           s.last_pass = atoi (optarg);
           if (s.last_pass < 1 || s.last_pass > 5)
             {
@@ -333,6 +448,10 @@ main (int argc, char * const argv [])
           s.include_path.push_back (string (optarg));
           break;
 
+        case 'd':
+          s.unwindsym_modules.insert (string (optarg));
+          break;
+
         case 'e':
 	  if (have_script)
 	    {
@@ -395,7 +514,6 @@ main (int argc, char * const argv [])
 	      }
 	  }
 
-	  cerr << "Warning: using '-m' disables cache support." << endl;
 	  s.use_cache = false;
           break;
 
@@ -406,6 +524,7 @@ main (int argc, char * const argv [])
 
         case 'k':
           s.keep_tmpdir = true;
+          s.use_cache = false; /* User wants to keep a usable build tree. */
           break;
 
         case 'g':
@@ -426,9 +545,9 @@ main (int argc, char * const argv [])
 
         case 's':
           s.buffer_size = atoi (optarg);
-          if (s.buffer_size < 1 || s.buffer_size > 64)
+          if (s.buffer_size < 1 || s.buffer_size > 4095)
             {
-              cerr << "Invalid buffer size (should be 1-64)." << endl;
+              cerr << "Invalid buffer size (should be 1-4095)." << endl;
 	      usage (s, 1);
             }
           break;
@@ -453,6 +572,51 @@ main (int argc, char * const argv [])
           usage (s, 0);
           break;
 
+        case 'l':
+	  s.suppress_warnings = true;
+          s.listing_mode = true;
+          s.last_pass = 2;
+          if (have_script)
+            {
+	      cerr << "Only one script can be given on the command line."
+		   << endl;
+	      usage (s, 1);
+            }
+          cmdline_script = string("probe ") + string(optarg) + " {}";
+          have_script = true;
+          break;
+
+        case 0:
+          switch (long_opt)
+            {
+            case LONG_OPT_KELF:
+	      s.consult_symtab = true;
+	      break;
+            case LONG_OPT_KMAP:
+	      // Leave s.consult_symtab unset for now, to ease error checking.
+              if (!s.kernel_symtab_path.empty())
+		{
+		  cerr << "You can't specify multiple --kmap options." << endl;
+		  usage(s, 1);
+		}
+              if (optarg)
+                s.kernel_symtab_path = optarg;
+              else
+#define PATH_TBD string("__TBD__")
+                s.kernel_symtab_path = PATH_TBD;
+	      break;
+	    case LONG_OPT_IGNORE_VMLINUX:
+	      s.ignore_vmlinux = true;
+	      break;
+	    case LONG_OPT_IGNORE_DWARF:
+	      s.ignore_dwarf = true;
+	      break;
+            default:
+              cerr << "Internal error parsing command arguments." << endl;
+              usage(s, 1);
+            }
+          break;
+
         default:
           usage (s, 1);
           break;
@@ -477,6 +641,18 @@ main (int argc, char * const argv [])
       usage (s, 1);
     }
 
+  if (!s.kernel_symtab_path.empty())
+    {
+      if (s.consult_symtab)
+      {
+        cerr << "You can't specify --kelf and --kmap together." << endl;
+	usage (s, 1);
+      }
+      s.consult_symtab = true;
+      if (s.kernel_symtab_path == PATH_TBD)
+        s.kernel_symtab_path = string("/boot/System.map-") + s.kernel_release;
+    }
+
   if (s.last_pass > 4 && release_changed)
     {
       if (s.verbose)
@@ -502,6 +678,16 @@ main (int argc, char * const argv [])
       usage(s, 1);
     }
 
+  // translate path of runtime to absolute path
+  if (s.runtime_path[0] != '/')
+    {
+      char cwd[PATH_MAX];
+      if (getcwd(cwd, sizeof(cwd)))
+        {
+          s.runtime_path = string(cwd) + "/" + s.runtime_path;
+        }
+    }
+
   int rc = 0;
 
   // override PATH and LC_ALL
@@ -528,7 +714,12 @@ main (int argc, char * const argv [])
 
   // arguments parsed; get down to business
   if (s.verbose > 1)
-    version ();
+    {
+      version ();
+      clog << "Session arch: " << s.architecture
+           << " release: " << s.kernel_release
+           << endl;
+    }
 
   // Create a temporary directory to build within.
   // Be careful with this, as "s.tmpdir" is "rm -rf"'d at the end.
@@ -561,10 +752,7 @@ main (int argc, char * const argv [])
 
   // Set up our handler to catch routine signals, to allow clean 
   // and reasonably timely exit.
-  signal (SIGHUP, handle_interrupt);
-  signal (SIGPIPE, handle_interrupt);
-  signal (SIGINT, handle_interrupt);
-  signal (SIGTERM, handle_interrupt);
+  setup_signals(&handle_interrupt);
 
   struct tms tms_before;
   times (& tms_before);
@@ -718,7 +906,7 @@ main (int argc, char * const argv [])
   // PASS 2: ELABORATION
   rc = semantic_pass (s);
 
-  if (rc == 0 && s.last_pass == 2)
+  if (s.listing_mode || (rc == 0 && s.last_pass == 2))
     printscript(s, cout);
 
   times (& tms_after);
@@ -767,7 +955,7 @@ main (int argc, char * const argv [])
 	}
     }
 
-  if (rc || s.last_pass == 2 || pending_interrupts) goto cleanup;
+  if (rc || s.listing_mode || s.last_pass == 2 || pending_interrupts) goto cleanup;
 
   // PASS 3: TRANSLATION
 
@@ -804,7 +992,10 @@ main (int argc, char * const argv [])
   rc = compile_pass (s);
 
   if (rc == 0 && s.last_pass == 4)
-    cout << s.hash_path << endl;
+    {
+      cout << ((s.hash_path == "") ? (s.module_name + string(".ko")) : s.hash_path);
+      cout << endl;
+    }
 
   times (& tms_after);
   gettimeofday (&tv_after, NULL);
@@ -885,6 +1076,10 @@ pass_5:
         clog << "Keeping temporary directory \"" << s.tmpdir << "\"" << endl;
       else
         {
+	  // Ignore signals while we're deleting the temporary directory.
+	  setup_signals (SIG_IGN);
+
+	  // Remove the temporary directory.
           string cleanupcmd = "rm -rf ";
           cleanupcmd += s.tmpdir;
           if (s.verbose>1) clog << "Running " << cleanupcmd << endl;
