@@ -103,8 +103,6 @@ extern char *__name__;
 #define STP_OLD_TRANSPORT
 #include "../transport/transport_msgs.h"
 
-extern int use_old_transport;
-
 #define RELAYFS_MAGIC	0xF0B4A981
 #define DEBUGFS_MAGIC	0x64626720
 #define DEBUGFSDIR	"/sys/kernel/debug"
@@ -118,9 +116,7 @@ int init_stapio(void);
 int stp_main_loop(void);
 int send_request(int type, void *data, int len);
 void cleanup_and_exit (int);
-int do_module(void *);
-int do_kernel_symbols(void);
-int init_ctl_channel(int);
+int init_ctl_channel(const char *name, int verb);
 void close_ctl_channel(void);
 int init_relayfs(void);
 void close_relayfs(void);
@@ -129,7 +125,7 @@ void close_oldrelayfs(int);
 void setup_signals(void);
 /* cap.c */
 void print_cap(char *text);
-int init_cap(void);
+void init_cap(void);
 void add_cap(cap_value_t cap);
 void del_cap(cap_value_t cap);
 void drop_cap(cap_value_t cap);
@@ -169,6 +165,7 @@ extern int target_pid;
 extern char *target_cmd;
 extern char *outfile_name;
 extern int attach_mod;
+extern int delete_mod;
 extern int load_only;
 extern int need_uprobes;
 
