@@ -25,7 +25,7 @@ unsigned long _stp_module_relocate(const char *module, const char *section, unsi
 	static struct _stp_module *last = NULL;
 	static struct _stp_symbol *last_sec;
 	unsigned long flags;
-	int i, j;
+	unsigned i, j;
 
 	/* if module is -1, we invalidate last. _stp_del_module calls this when modules are deleted. */
 	if ((long)module == -1) {
@@ -33,12 +33,10 @@ unsigned long _stp_module_relocate(const char *module, const char *section, unsi
 		return 0;
 	}
 
-	dbug(DEBUG_SYMBOLS, "%s, %s, %lx\n", module, section, offset);
+	dbug_sym(1, "%s, %s, %lx\n", module, section, offset);
 
-	STP_RLOCK_MODULES;
 	if (!module || !strcmp(section, "")	/* absolute, unrelocated address */
 	    ||_stp_num_modules == 0) {
-		STP_RUNLOCK_MODULES;
 		return offset;
 	}
 
@@ -46,119 +44,80 @@ unsigned long _stp_module_relocate(const char *module, const char *section, unsi
 	if (last) {
 		if (!strcmp(module, last->name) && !strcmp(section, last_sec->symbol)) {
 			offset += last_sec->addr;
-			STP_RUNLOCK_MODULES;
-			dbug(DEBUG_SYMBOLS, "offset = %lx\n", offset);
+			dbug_sym(1, "cached address=%lx\n", offset);
 			return offset;
 		}
 	}
 
-	/* not cached. need to scan all modules */
-	if (!strcmp(module, "kernel")) {
-		STP_RUNLOCK_MODULES;
+        for (i = 0; i < _stp_num_modules; i++) {
+          last = _stp_modules[i];
+          if (strcmp(module, last->name))
+            continue;
+          for (j = 0; j < last->num_sections; j++) {
+            last_sec = &last->sections[j];
+            if (!strcmp(section, last_sec->symbol)) {
+            
+              if (last_sec->addr == 0) /* module/section not in memory */
+                continue;
 
-		/* See also transport/symbols.c (_stp_do_symbols). */
-		if (strcmp(section, "_stext"))
-			return 0;
-		else
-			return offset + _stp_modules[0]->text;
-	} else {
-		/* relocatable module */
-		for (i = 1; i < _stp_num_modules; i++) {	/* skip over [0]=kernel */
-			last = _stp_modules[i];
-			if (strcmp(module, last->name))
-				continue;
-			for (j = 0; j < (int)last->num_sections; j++) {
-				last_sec = &last->sections[j];
-				if (!strcmp(section, last_sec->symbol)) {
-					offset += last_sec->addr;
-					STP_RUNLOCK_MODULES;
-					dbug(DEBUG_SYMBOLS, "offset = %lx\n", offset);
-					return offset;
-				}
-			}
-		}
+              offset += last_sec->addr;
+              dbug_sym(1, "address=%lx\n", offset);
+              return offset;
+            }
+          }
 	}
-	STP_RUNLOCK_MODULES;
+
 	last = NULL;
 	return 0;
 }
 
-/* Lookup the kernel address for this symbol. Returns 0 if not found. */
-static unsigned long _stp_kallsyms_lookup_name(const char *name)
-{
-	struct _stp_symbol *s = _stp_modules[0]->symbols;
-	unsigned num = _stp_modules[0]->num_symbols;
-
-	while (num--) {
-		if (strcmp(name, s->symbol) == 0)
-			return s->addr;
-		s++;
-	}
-	return 0;
-}
 
+/* Return the module that likely contains the given address.  */
+/* XXX: This query only makes sense with respect to a particular
+   address space.  A more general interface would have to identify
+   the address space, and also pass back the section. */ 
 static struct _stp_module *_stp_find_module_by_addr(unsigned long addr)
 {
-	unsigned begin = 0;
-	unsigned end = _stp_num_modules;
+  unsigned i;
+  struct _stp_module *closest_module = NULL;
+  unsigned long closest_module_offset = ~0; /* minimum[addr - module->.text] */
+  
+  for (i=0; i<_stp_num_modules; i++)
+    {
+      unsigned long module_text_addr, this_module_offset;
 
-	if (unlikely(addr < _stp_modules_by_addr[0]->text))
-		return NULL;
+      if (_stp_modules[i]->num_sections < 1) continue;
+      module_text_addr = _stp_modules[i]->sections[0].addr; /* XXX: assume section[0]=>text */
+      if (addr < module_text_addr) continue;
+      this_module_offset = module_text_addr - addr;
+      
+      if (this_module_offset < closest_module_offset)
+        {
+          closest_module = _stp_modules[i];
+          closest_module_offset = this_module_offset;
+        }
+    }
 
-	if (_stp_num_modules > 1 && addr > _stp_modules_by_addr[0]->data) {
-		/* binary search on index [begin,end) */
-		do {
-			unsigned mid = (begin + end) / 2;
-			if (addr < _stp_modules_by_addr[mid]->text)
-				end = mid;
-			else
-				begin = mid;
-		} while (begin + 1 < end);
-		/* result index in $begin, guaranteed between [0,_stp_num_modules) */
-	}
-	/* check if addr is past the last module */
-	if (unlikely(begin == _stp_num_modules - 1
-		     && (addr > _stp_modules_by_addr[begin]->text + _stp_modules_by_addr[begin]->text_size)))
-		return NULL;
-
-	return _stp_modules_by_addr[begin];
+  return closest_module;
 }
 
-static struct _stp_module *_stp_get_unwind_info(unsigned long addr)
-{
-	struct _stp_module *m;
-	struct _stp_symbol *s;
-	unsigned long flags;
 
-	STP_RLOCK_MODULES;
-	m = _stp_find_module_by_addr(addr);
-	if (unlikely(m == NULL)) {
-		STP_RUNLOCK_MODULES;
-		return NULL;
-	}
-	/* Lock the module struct so it doesn't go away while being used. */
-	/* Probably could never happen, but lock it to be sure for now. */
-	read_lock(&m->lock);
 
-	STP_RUNLOCK_MODULES;
-	return m;
-}
-
-static const char *_stp_kallsyms_lookup(unsigned long addr,
-					unsigned long *symbolsize, unsigned long *offset, char **modname, char *namebuf)
+static const char *_stp_kallsyms_lookup(unsigned long addr, unsigned long *symbolsize,
+                                        unsigned long *offset, char **modname, char *namebuf)
 {
 	struct _stp_module *m;
 	struct _stp_symbol *s;
 	unsigned long flags;
 	unsigned end, begin = 0;
 
-	STP_RLOCK_MODULES;
 	m = _stp_find_module_by_addr(addr);
 	if (unlikely(m == NULL)) {
-		STP_RUNLOCK_MODULES;
 		return NULL;
 	}
 
+        /* NB: relativize the address to the (XXX) presumed text section. */
+        addr -= m->sections[0].addr;
 	end = m->num_symbols;
 
 	/* binary search for symbols within the module */
@@ -187,17 +146,15 @@ static const char *_stp_kallsyms_lookup(unsigned long addr,
 		}
 		if (namebuf) {
 			strlcpy(namebuf, s->symbol, KSYM_NAME_LEN + 1);
-			STP_RUNLOCK_MODULES;
 			return namebuf;
 		} else {
-			STP_RUNLOCK_MODULES;
 			return s->symbol;
 		}
 	}
-	STP_RUNLOCK_MODULES;
 	return NULL;
 }
 
+
 /** Print an address symbolically.
  * @param address The address to lookup.
  * @note Symbolic lookups should not normally be done within
@@ -223,8 +180,7 @@ void _stp_symbol_print(unsigned long address)
 }
 
 /* Like _stp_symbol_print, except only print if the address is a valid function address */
-
-void _stp_func_print(unsigned long address, int verbose, int exact)
+int _stp_func_print(unsigned long address, int verbose, int exact)
 {
 	char *modname;
 	const char *name;
@@ -247,7 +203,9 @@ void _stp_func_print(unsigned long address, int verbose, int exact)
 				_stp_printf(" %p : %s+%#lx/%#lx%s\n", (int64_t) address, name, offset, size, exstr);
 		} else
 			_stp_printf("%p ", (int64_t) address);
+		return 1;
 	}
+	return 0;
 }
 
 void _stp_symbol_snprint(char *str, size_t len, unsigned long address)
